import asyncio
import hashlib
import json
import logging
import time
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

import aiohttp
from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.wallet import LocalWallet
from cosmpy.crypto.address import Address
from pydantic import BaseModel

from uagents.config import (
    ALMANAC_API_MAX_RETRIES,
    ALMANAC_API_TIMEOUT_SECONDS,
    ALMANAC_API_URL,
    ALMANAC_CONTRACT_VERSION,
    ALMANAC_REGISTRATION_WAIT,
    REGISTRATION_FEE,
    REGISTRATION_UPDATE_INTERVAL_SECONDS,
)
from uagents.crypto import Identity
from uagents.network import (
    AlmanacContract,
    AlmanacContractRecord,
    InsufficientFundsError,
    add_testnet_funds,
)
from uagents.types import AgentEndpoint, AgentInfo


class VerifiableModel(BaseModel):
    agent_address: str
    signature: Optional[str] = None
    timestamp: Optional[int] = None

    def sign(self, identity: Identity):
        self.timestamp = int(time.time())
        digest = self._build_digest()
        self.signature = identity.sign_digest(digest)

    def verify(self) -> bool:
        return self.signature is not None and Identity.verify_digest(
            self.agent_address, self._build_digest(), self.signature
        )

    def _build_digest(self) -> bytes:
        sha256 = hashlib.sha256()
        sha256.update(
            json.dumps(
                self.model_dump(exclude={"signature"}),
                sort_keys=True,
                separators=(",", ":"),
            ).encode("utf-8")
        )
        return sha256.digest()


class AgentRegistrationAttestation(VerifiableModel):
    protocols: List[str]
    endpoints: List[AgentEndpoint]
    metadata: Optional[Dict[str, Union[str, Dict[str, str]]]] = None


class AgentRegistrationAttestationBatch(BaseModel):
    attestations: List[AgentRegistrationAttestation]


class AgentStatusUpdate(VerifiableModel):
    is_active: bool


def generate_backoff_time(retry: int) -> float:
    """
    Generate a backoff time starting from 0.128 seconds and limited to ~131 seconds
    """
    return (2 ** (min(retry, 11) + 6)) / 1000


def coerce_metadata_to_str(
    metadata: Optional[Dict[str, Any]],
) -> Optional[Dict[str, Union[str, Dict[str, str]]]]:
    """
    Step through the metadata and convert any non-string values to strings.
    """
    if metadata is None:
        return None
    out = {}
    for key, val in metadata.items():
        if isinstance(val, dict):
            out[key] = {
                k: v if isinstance(v, str) else json.dumps(v) for k, v in val.items()
            }
        else:
            out[key] = val if isinstance(val, str) else json.dumps(val)
    return out


def extract_geo_metadata(
    metadata: Optional[Dict[str, Any]],
) -> Optional[Dict[str, Any]]:
    """
    Extract geo-location metadata from the metadata dictionary.
    """
    if metadata is None:
        return None
    return {k: v for k, v in metadata.items() if k == "geolocation"}


async def almanac_api_post(
    url: str, data: BaseModel, raise_from: bool = True, retries: int = 3
) -> bool:
    async with aiohttp.ClientSession() as session:
        for retry in range(retries):
            try:
                async with session.post(
                    url,
                    headers={"content-type": "application/json"},
                    data=data.model_dump_json(),
                    timeout=aiohttp.ClientTimeout(total=ALMANAC_API_TIMEOUT_SECONDS),
                ) as resp:
                    resp.raise_for_status()
                    return True
            except (aiohttp.ClientError, asyncio.exceptions.TimeoutError) as e:
                if retry == retries - 1:
                    if raise_from:
                        raise e
                    return False
                await asyncio.sleep(generate_backoff_time(retry))
    return False


class AgentRegistrationPolicy(ABC):
    @abstractmethod
    # pylint: disable=unnecessary-pass
    async def register(
        self,
        agent_address: str,
        protocols: List[str],
        endpoints: List[AgentEndpoint],
        metadata: Optional[Dict[str, Any]] = None,
    ):
        pass


class BatchRegistrationPolicy(ABC):
    @abstractmethod
    # pylint: disable=unnecessary-pass
    async def register(self):
        pass

    @abstractmethod
    def add_agent(self, agent_info: AgentInfo, identity: Identity):
        pass


class AlmanacApiRegistrationPolicy(AgentRegistrationPolicy):
    def __init__(
        self,
        identity: Identity,
        *,
        almanac_api: Optional[str] = None,
        max_retries: int = ALMANAC_API_MAX_RETRIES,
        logger: Optional[logging.Logger] = None,
    ):
        self._almanac_api = almanac_api or ALMANAC_API_URL
        self._max_retries = max_retries
        self._identity = identity
        self._logger = logger or logging.getLogger(__name__)

    async def register(
        self,
        agent_address: str,
        protocols: List[str],
        endpoints: List[AgentEndpoint],
        metadata: Optional[Dict[str, Any]] = None,
    ):
        # create the attestation
        attestation = AgentRegistrationAttestation(
            agent_address=agent_address,
            protocols=protocols,
            endpoints=endpoints,
            metadata=coerce_metadata_to_str(extract_geo_metadata(metadata)),
        )

        # sign the attestation
        attestation.sign(self._identity)

        success = await almanac_api_post(
            f"{self._almanac_api}/agents", attestation, retries=self._max_retries
        )
        if success:
            self._logger.info("Registration on Almanac API successful")
        else:
            self._logger.warning("Registration on Almanac API failed")


class BatchAlmanacApiRegistrationPolicy(AgentRegistrationPolicy):
    def __init__(
        self, almanac_api: Optional[str] = None, logger: Optional[logging.Logger] = None
    ):
        self._almanac_api = almanac_api or ALMANAC_API_URL
        self._attestations: List[AgentRegistrationAttestation] = []
        self._logger = logger or logging.getLogger(__name__)

    def add_agent(self, agent_info: AgentInfo, identity: Identity):
        attestation = AgentRegistrationAttestation(
            agent_address=agent_info.agent_address,
            protocols=list(agent_info.protocols),
            endpoints=agent_info.endpoints,
            metadata=coerce_metadata_to_str(extract_geo_metadata(agent_info.metadata)),
        )
        attestation.sign(identity)
        self._attestations.append(attestation)

    async def register(self):
        if not self._attestations:
            return
        attestations = AgentRegistrationAttestationBatch(
            attestations=self._attestations
        )
        success = await almanac_api_post(
            f"{self._almanac_api}/agents/batch", attestations
        )
        if success:
            self._logger.info("Batch registration on Almanac API successful")
        else:
            self._logger.warning("Batch registration on Almanac API failed")


class LedgerBasedRegistrationPolicy(AgentRegistrationPolicy):
    def __init__(
        self,
        identity: Identity,
        ledger: LedgerClient,
        wallet: LocalWallet,
        almanac_contract: AlmanacContract,
        testnet: bool,
        *,
        logger: Optional[logging.Logger] = None,
    ):
        self._identity = identity
        self._wallet = wallet
        self._ledger = ledger
        self._testnet = testnet
        self._almanac_contract = almanac_contract
        self._logger = logger or logging.getLogger(__name__)

    def check_contract_version(self):
        """
        Check the version of the deployed Almanac contract and log a warning
        if it is different from the supported version.
        """
        deployed_version = self._almanac_contract.get_contract_version()
        if deployed_version != ALMANAC_CONTRACT_VERSION:
            self._logger.warning(
                "Mismatch in almanac contract versions: supported (%s), deployed (%s). "
                "Update uAgents to the latest version for compatibility.",
                ALMANAC_CONTRACT_VERSION,
                deployed_version,
            )

    async def register(
        self,
        agent_address: str,
        protocols: List[str],
        endpoints: List[AgentEndpoint],
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """
        Register the agent on the Almanac contract if registration is about to expire or
        the registration data has changed.
        """
        self.check_contract_version()

        if (
            not self._almanac_contract.is_registered(agent_address)
            or self._almanac_contract.get_expiry(agent_address)
            < REGISTRATION_UPDATE_INTERVAL_SECONDS
            or endpoints != self._almanac_contract.get_endpoints(agent_address)
            or protocols != self._almanac_contract.get_protocols(agent_address)
        ):
            if self._get_balance() < REGISTRATION_FEE:
                self._logger.warning(
                    "I do not have enough funds to register on Almanac contract"
                )
                if self._testnet:
                    add_testnet_funds(str(self._wallet.address()))
                    self._logger.info(
                        f"Adding testnet funds to {self._wallet.address()}"
                    )
                else:
                    self._logger.info(
                        f"Send funds to wallet address: {self._wallet.address()}"
                    )
                raise InsufficientFundsError()

            self._logger.info("Registering on almanac contract...")

            current_time = int(time.time()) - ALMANAC_REGISTRATION_WAIT

            signature = self._sign_registration(current_time)
            await self._almanac_contract.register(
                self._ledger,
                self._wallet,
                agent_address,
                protocols,
                endpoints,
                signature,
                current_time,
            )
            self._logger.info("Registering on almanac contract...complete")
        else:
            self._logger.info("Almanac contract registration is up to date!")

    def _get_balance(self) -> int:
        return self._ledger.query_bank_balance(Address(self._wallet.address()))

    def _sign_registration(self, timestamp: int) -> str:
        """
        Sign the registration data for Almanac contract.

        Args:
            timestamp (int): The timestamp for the registration.

        Returns:
            str: The signature of the registration data.

        Raises:
            AssertionError: If the Almanac contract address is None.

        """
        assert self._almanac_contract.address is not None
        return self._identity.sign_registration(
            str(self._almanac_contract.address),
            timestamp,
            str(self._wallet.address()),
        )


class BatchLedgerRegistrationPolicy(BatchRegistrationPolicy):
    def __init__(
        self,
        ledger: LedgerClient,
        wallet: LocalWallet,
        almanac_contract: AlmanacContract,
        testnet: bool,
        logger: Optional[logging.Logger] = None,
    ):
        self._ledger = ledger
        self._wallet = wallet
        self._almanac_contract = almanac_contract
        self._testnet = testnet
        self._logger = logger or logging.getLogger(__name__)
        self._records: List[AlmanacContractRecord] = []
        self._identities: Dict[str, Identity] = {}

    def add_agent(self, agent_info: AgentInfo, identity: Identity):
        agent_record = AlmanacContractRecord(
            agent_address=agent_info.agent_address,
            protocols=agent_info.protocols,
            endpoints=agent_info.endpoints,
            contract_address=str(self._almanac_contract.address),
            sender_address=str(self._wallet.address()),
        )
        self._records.append(agent_record)
        self._identities[agent_info.agent_address] = identity

    def _get_balance(self) -> int:
        return self._ledger.query_bank_balance(Address(self._wallet.address()))

    async def register(self):
        self._logger.info("Registering agents on Almanac contract...")
        for record in self._records:
            record.sign(self._identities[record.agent_address])

        if self._get_balance() < REGISTRATION_FEE * len(self._records):
            self._logger.warning(
                f"I do not have enough funds to register {len(self._records)} "
                "agents on Almanac contract"
            )
            if self._testnet:
                add_testnet_funds(str(self._wallet.address()))
                self._logger.info(f"Adding testnet funds to {self._wallet.address()}")
            else:
                self._logger.info(
                    f"Send funds to wallet address: {self._wallet.address()}"
                )
            raise InsufficientFundsError()

        await self._almanac_contract.register_batch(
            self._ledger, self._wallet, self._records
        )

        self._logger.info("Registering agents on Almanac contract...complete")


class DefaultRegistrationPolicy(AgentRegistrationPolicy):
    def __init__(
        self,
        identity: Identity,
        ledger: LedgerClient,
        wallet: LocalWallet,
        almanac_contract: Optional[AlmanacContract],
        testnet: bool,
        *,
        logger: Optional[logging.Logger] = None,
        almanac_api: Optional[str] = None,
    ):
        self._logger = logger or logging.getLogger(__name__)
        self._api_policy = AlmanacApiRegistrationPolicy(
            identity, almanac_api=almanac_api, logger=logger
        )
        if almanac_contract is None:
            self._ledger_policy = None
        else:
            self._ledger_policy = LedgerBasedRegistrationPolicy(
                identity, ledger, wallet, almanac_contract, testnet, logger=logger
            )

    async def register(
        self,
        agent_address: str,
        protocols: List[str],
        endpoints: List[AgentEndpoint],
        metadata: Optional[Dict[str, Any]] = None,
    ):
        # prefer the API registration policy as it is faster
        try:
            await self._api_policy.register(
                agent_address, protocols, endpoints, metadata
            )
        except Exception as e:
            self._logger.warning(
                f"Failed to register on Almanac API: {e.__class__.__name__}"
            )

        if self._ledger_policy is None:
            return

        # schedule the ledger registration
        try:
            await self._ledger_policy.register(
                agent_address, protocols, endpoints, metadata
            )
        except InsufficientFundsError:
            self._logger.warning(
                "Failed to register on Almanac contract due to insufficient funds"
            )
            raise
        except Exception as e:
            self._logger.error(f"Failed to register on Almanac contract: {e}")
            raise


async def update_agent_status(status: AgentStatusUpdate, almanac_api: str):
    await almanac_api_post(
        f"{almanac_api}/agents/{status.agent_address}/status",
        status,
        raise_from=False,
    )


class DefaultBatchRegistrationPolicy(BatchRegistrationPolicy):
    def __init__(
        self,
        ledger: LedgerClient,
        wallet: Optional[LocalWallet] = None,
        almanac_contract: Optional[AlmanacContract] = None,
        testnet: bool = True,
        *,
        logger: Optional[logging.Logger] = None,
        almanac_api: Optional[str] = None,
    ):
        self._logger = logger or logging.getLogger(__name__)
        self._api_policy = BatchAlmanacApiRegistrationPolicy(
            almanac_api=almanac_api, logger=logger
        )

        if almanac_contract is None or wallet is None:
            self._ledger_policy = None
        else:
            self._ledger_policy = BatchLedgerRegistrationPolicy(
                ledger, wallet, almanac_contract, testnet, logger=logger
            )

    def add_agent(self, agent_info: AgentInfo, identity: Identity):
        self._api_policy.add_agent(agent_info, identity)
        if self._ledger_policy is not None:
            self._ledger_policy.add_agent(agent_info, identity)

    async def register(self):
        # prefer the API registration policy as it is faster
        try:
            await self._api_policy.register()
        except Exception as e:
            self._logger.warning(
                f"Failed to batch register on Almanac API: {e.__class__.__name__}"
            )

        if self._ledger_policy is None:
            return

        # schedule the ledger registration
        try:
            await self._ledger_policy.register()
        except InsufficientFundsError:
            self._logger.warning(
                "Failed to batch register on Almanac contract due to insufficient funds"
            )
            raise
        except Exception as e:
            self._logger.error(f"Failed to batch register on Almanac contract: {e}")
            raise



import hashlib
from typing import Any, Dict, Type, Union

from pydantic.v1 import BaseModel, Field  # noqa


# reverting back to pydantic v1 BaseModel for backwards compatibility
class Model(BaseModel):
    @classmethod
    def model_json_schema(cls) -> str:
        return cls.schema_json()

    def model_dump_json(self) -> str:
        return self.json()

    def model_dump(self) -> Dict[str, Any]:
        return self.dict()

    @classmethod
    def model_validate_json(cls, obj: Any) -> "Model":
        return cls.parse_raw(obj)

    @classmethod
    def model_validate(cls, obj: Union[Dict[str, Any], "Model"]) -> "Model":
        return cls.parse_obj(obj)

    @staticmethod
    def build_schema_digest(model: Union["Model", Type["Model"]]) -> str:
        schema = model.schema_json(indent=None, sort_keys=True)
        digest = hashlib.sha256(schema.encode("utf8")).digest().hex()

        return f"model:{digest}"


class ErrorMessage(Model):
    """Error message model"""

    error: str



"""
This Protocol class can be used to rate limit `on_message` message handlers.

The rate limiter uses the agents storage to keep track of the number of requests
made by another agent within a given time window. If the number of requests exceeds
a specified limit, the rate limiter will block further requests until the time
window resets.

> Default: Not rate limited, but you can set a default during initialization.

Additionally, the protocol can be used to set access control rules for handlers
allowing or blocking specific agents from accessing the handler.
The default access control rule can be set to allow or block all agents.

Both rules can work together to provide a secure and rate-limited environment for
message handlers.


Usage examples:

```python
from uagents.experimental.quota import AccessControlList, QuotaProtocol, RateLimit

# Initialize the QuotaProtocol instance
quota_protocol = QuotaProtocol(
    storage_reference=agent.storage,
    name="quota_proto",
    version=agent._version,
    # default_rate_limit=RateLimit(window_size_minutes=1, max_requests=3), # Optional
)

# This message handler is not rate limited
@quota_protocol.on_message(ExampleMessage1)
async def handle(ctx: Context, sender: str, msg: ExampleMessage1):
    ...

# This message handler is rate limited with custom window size and request limit
@quota_protocol.on_message(
    ExampleMessage2,
    rate_limit=RateLimit(window_size_minutes=1, max_requests=3),
)
async def handle(ctx: Context, sender: str, msg: ExampleMessage2):
    ...

# This message handler has access control rules set
@quota_protocol.on_message(
    ExampleMessage3,
    acl=AccessControlList(default=False, allowed={"<agent_address>"}),
)
async def handle(ctx: Context, sender: str, msg: ExampleMessage3):
    ...

agent.include(quota_protocol)
```

Tip: The `AccessControlList` object can be used to set access control rules during
runtime. This can be useful for dynamic access control rules based on the state of the
agent or the network.
```python
acl = AccessControlList(default=True)

@proto.on_message(model=Message, access_control_list=acl)
async def message_handler(ctx: Context, sender: str, msg: Message):
    if REASON_TO_BLOCK:
        acl.blocked.add(sender)
    ctx.logger.info(f"Received message from {sender}: {msg.text}")
```

"""

import functools
import time
from typing import Optional, Set, Type, Union

from pydantic import BaseModel

from uagents import Context, Model, Protocol
from uagents.models import ErrorMessage
from uagents.storage import StorageAPI
from uagents.types import MessageCallback

WINDOW_SIZE_MINUTES = 60
MAX_REQUESTS = 6


class Usage(BaseModel):
    time_window_start: float
    window_size_minutes: int
    requests: int
    max_requests: int


class RateLimit(BaseModel):
    window_size_minutes: int
    max_requests: int


class AccessControlList(BaseModel):
    default: bool
    allowed: set[str] = set()
    blocked: set[str] = set()
    bypass_rate_limit: set[str] = set()


class QuotaProtocol(Protocol):
    def __init__(
        self,
        storage_reference: StorageAPI,
        name: Optional[str] = None,
        version: Optional[str] = None,
        default_rate_limit: Optional[RateLimit] = None,
        default_acl: Optional[AccessControlList] = None,
    ):
        """
        Initialize a QuotaProtocol instance.

        Args:
            storage_reference (StorageAPI): The storage reference to use for rate limiting.
            name (Optional[str], optional): The name of the protocol. Defaults to None.
            version (Optional[str], optional): The version of the protocol. Defaults to None.
            default_rate_limit (Optional[RateLimit], optional): The default rate limit.
                Defaults to None.
            default_acl (Optional[AccessControlList], optional): The access control list.
                Defaults to None.
        """
        super().__init__(name=name, version=version)
        self.storage_ref = storage_reference
        self.default_rate_limit = default_rate_limit
        self.default_acl = default_acl

    def on_message(
        self,
        model: Type[Model],
        replies: Optional[Union[Type[Model], Set[Type[Model]]]] = None,
        allow_unverified: Optional[bool] = False,
        rate_limit: Optional[RateLimit] = None,
        access_control_list: Optional[AccessControlList] = None,
    ):
        """
        Overwritten decorator to register a message handler for the protocol
        including rate limiting.

        Args:
            model (Type[Model]): The message model type.
            replies (Optional[Union[Type[Model], Set[Type[Model]]]], optional): The associated
            reply types. Defaults to None.
            allow_unverified (Optional[bool], optional): Whether to allow unverified messages.
            Defaults to False.
            rate_limit (Optional[RateLimit], optional): The rate limit to apply. Defaults to None.
            access_control_list (Optional[AccessControlList], optional): The access control list to
            apply.

        Returns:
            Callable: The decorator to register the message handler.
        """

        def decorator_on_message(func: MessageCallback):
            handler = self.wrap(func, rate_limit, access_control_list)
            self._add_message_handler(model, handler, replies, allow_unverified)
            return handler

        return decorator_on_message

    def wrap(
        self,
        func: MessageCallback,
        rate_limit: Optional[RateLimit] = None,
        acl: Optional[AccessControlList] = None,
    ) -> MessageCallback:
        """
        Decorator to wrap a function with rate limiting.

        Args:
            func: The function to wrap with rate limiting
            rate_limit: The rate limit to apply
            acl: The access control list to apply

        Returns:
            Callable: The decorated
        """
        acl = acl or self.default_acl
        if acl is None:
            acl = AccessControlList(default=True)

        rate_limit = rate_limit or self.default_rate_limit

        @functools.wraps(func)
        async def decorator(ctx: Context, sender: str, msg: Type[Model]):
            if (acl.default and sender in acl.blocked) or (
                not acl.default and sender not in acl.allowed
            ):
                return await ctx.send(
                    sender,
                    ErrorMessage(error=("You are not allowed to access this handler.")),
                )
            if (
                sender in acl.bypass_rate_limit
                or not rate_limit
                or self.add_request(
                    sender,
                    func.__name__,
                    rate_limit.window_size_minutes,
                    rate_limit.max_requests,
                )
            ):
                result = await func(ctx, sender, msg)
            else:
                err = (
                    f"Rate limit exceeded for {msg.schema()['title']}. "
                    f"This handler allows for {rate_limit.max_requests} calls per "
                    f"{rate_limit.window_size_minutes} minutes. Try again later."
                )
                result = await ctx.send(sender, ErrorMessage(error=err))
            return result

        return decorator  # type: ignore

    def _clean_usage(self, usage: dict[str, dict]):
        """
        Remove all time windows that are older than the current time window.

        Args:
            usage: The usage dictionary to clean
        """
        now = int(time.time())
        for key in list(usage.keys()):
            if (now - usage[key]["time_window_start"]) > usage[key][
                "window_size_minutes"
            ] * 60:
                del usage[key]

    def add_request(
        self,
        agent_address: str,
        function_name: str,
        window_size_minutes: int,
        max_requests: int,
    ) -> bool:
        """
        Add a request to the rate limiter if the current time is still within the
        time window since the beginning of the most recent time window. Otherwise,
        reset the time window and add the request.

        Args:
            agent_address: The address of the agent making the request

        Returns:
            False if the maximum number of requests has been exceeded, True otherwise
        """

        now = int(time.time())

        usage = self.storage_ref.get(agent_address) or {}

        if function_name in usage:
            quota = Usage(**usage[function_name])
            if (now - quota.time_window_start) <= window_size_minutes * 60:
                if quota.requests >= max_requests:
                    return False
                quota.requests += 1
            else:
                quota.time_window_start = now
                quota.requests = 1
            usage[function_name] = quota.model_dump()
        else:
            usage[function_name] = Usage(
                time_window_start=now,
                window_size_minutes=window_size_minutes,
                requests=1,
                max_requests=max_requests,
            ).model_dump()

        self._clean_usage(usage)

        self.storage_ref.set(agent_address, usage)

        return True



"""Agent dispatch of exchange envelopes and synchronous messages."""

import asyncio
import logging
import uuid
from time import time
from typing import List, Optional, Tuple, Type, Union

import aiohttp
from pydantic import UUID4, ValidationError

from uagents.config import DEFAULT_ENVELOPE_TIMEOUT_SECONDS
from uagents.crypto import Identity, is_user_address
from uagents.dispatch import dispatcher
from uagents.envelope import Envelope, EnvelopeHistory, EnvelopeHistoryEntry
from uagents.models import Model
from uagents.resolver import GlobalResolver, Resolver
from uagents.types import DeliveryStatus, JsonStr, MsgStatus
from uagents.utils import get_logger

LOGGER = get_logger("dispenser", logging.DEBUG)


class Dispenser:
    """
    Dispenses messages externally.
    """

    def __init__(self, msg_cache_ref: Optional[EnvelopeHistory] = None):
        self._envelopes: asyncio.Queue[
            Tuple[Envelope, List[str], asyncio.Future, bool]
        ] = asyncio.Queue()
        self._msg_cache_ref = msg_cache_ref

    def add_envelope(
        self,
        envelope: Envelope,
        endpoints: List[str],
        response_future: asyncio.Future,
        sync: bool = False,
    ):
        """
        Add an envelope to the dispenser.

        Args:
            envelope (Envelope): The envelope to send.
            endpoints (List[str]): The endpoints to send the envelope to.
            response_future (asyncio.Future): The future to set the response on.
            sync (bool, optional): True if the message is synchronous. Defaults to False.
        """
        self._envelopes.put_nowait((envelope, endpoints, response_future, sync))

    async def run(self):
        """Run the dispenser routine."""
        while True:
            # get the message from the queue
            env, endpoints, response_future, sync = await self._envelopes.get()

            try:
                result = await send_exchange_envelope(
                    envelope=env,
                    endpoints=endpoints,
                    sync=sync,
                )
                response_future.set_result(result)

                if self._msg_cache_ref:
                    self._msg_cache_ref.add_entry(
                        EnvelopeHistoryEntry.from_envelope(env)
                    )
            except Exception as err:
                LOGGER.error(f"Failed to send envelope: {err}")


async def dispatch_local_message(
    sender: str,
    destination: str,
    schema_digest: str,
    message: JsonStr,
    session_id: uuid.UUID,
) -> MsgStatus:
    """Process a message locally."""
    await dispatcher.dispatch_msg(
        sender=sender,
        destination=destination,
        schema_digest=schema_digest,
        message=message,
        session=session_id,
    )
    return MsgStatus(
        status=DeliveryStatus.DELIVERED,
        detail="Message dispatched locally",
        destination=destination,
        endpoint="",
        session=session_id,
    )


async def send_exchange_envelope(
    envelope: Envelope,
    endpoints: List[str],
    sync: bool = False,
) -> Union[MsgStatus, Envelope]:
    """
    Method to send an exchange envelope.

    Args:
        envelope (Envelope): The envelope to send.
        resolver (Optional[Resolver], optional): The resolver to use. Defaults to None.
        sync (bool, optional): True if the message is synchronous. Defaults to False.

    Returns:
        Union[MsgStatus, Envelope]: Either the status of the message or the response envelope.
    """
    headers = {"content-type": "application/json"}
    if sync:
        headers["x-uagents-connection"] = "sync"
    errors = []
    for endpoint in endpoints:
        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    endpoint,
                    headers=headers,
                    data=envelope.model_dump_json(),
                ) as resp:
                    success = resp.status == 200
                    if success:
                        if sync:
                            env = Envelope.model_validate(await resp.json())
                            if env.signature:
                                verified = False
                                try:
                                    verified = env.verify()
                                except Exception as ex:
                                    errors.append(
                                        f"Received response envelope that failed verification: {ex}"
                                    )
                                if not verified:
                                    continue
                            return await dispatch_sync_response_envelope(env)
                        return MsgStatus(
                            status=DeliveryStatus.DELIVERED,
                            detail="Message successfully delivered via HTTP",
                            destination=envelope.target,
                            endpoint=endpoint,
                            session=envelope.session,
                        )
                errors.append(await resp.text())
        except aiohttp.ClientConnectorError as ex:
            errors.append(f"Failed to connect: {ex}")
        except ValidationError as ex:
            errors.append(f"Invalid sync response: {ex}")
        except Exception as ex:
            errors.append(f"Failed to send message: {ex}")
    LOGGER.error(
        f"Failed to deliver message to {envelope.target} @ {endpoints}: " + str(errors)
    )
    return MsgStatus(
        status=DeliveryStatus.FAILED,
        detail="Message delivery failed",
        destination=envelope.target,
        endpoint="",
        session=envelope.session,
    )


async def dispatch_sync_response_envelope(env: Envelope) -> Union[MsgStatus, Envelope]:
    """Dispatch a synchronous response envelope locally."""
    # If there are no sinks registered, return the envelope back to the caller
    if len(dispatcher.sinks) == 0:
        return env
    await dispatcher.dispatch_msg(
        env.sender,
        env.target,
        env.schema_digest,
        env.decode_payload(),
        env.session,
    )
    return MsgStatus(
        status=DeliveryStatus.DELIVERED,
        detail="Sync message successfully delivered via HTTP",
        destination=env.target,
        endpoint="",
        session=env.session,
    )


async def send_message_raw(
    destination: str,
    message_schema_digest: str,
    message_body: JsonStr,
    response_type: Optional[Type[Model]] = None,
    sender: Optional[Union[Identity, str]] = None,
    resolver: Optional[Resolver] = None,
    timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    sync: bool = False,
) -> Union[Model, JsonStr, MsgStatus, Envelope]:
    """
    Standalone function to send a message to an agent.

    Args:
        destination (str): The destination address to send the message to.
        message_schema_digest (str): The schema digest of the message.
        message_body (JsonStr): The JSON-formatted message to be sent.
        response_type (Optional[Type[Model]]): The optional type of the response message.
        sender (Optional[Union[Identity, str]]): The optional sender identity or user address.
        resolver (Optional[Resolver]): The optional resolver for address-to-endpoint resolution.
        timeout (int): The timeout for the message response in seconds. Defaults to 30.
        sync (bool): True if the message is synchronous.

    Returns:
        Union[Model, JsonStr, MsgStatus, Envelope]: On success, if the response type is provided,
        the response message is returned with that type. Otherwise, the JSON message is returned.
        If the sender is a user address, the response envelope is returned.
        On failure, a message status is returned.
    """
    if isinstance(sender, str) and is_user_address(sender):
        sender_address = sender
    if sender is None:
        sender = Identity.generate()
    if isinstance(sender, Identity):
        sender_address = sender.address
    if not sender_address:
        raise ValueError("Invalid sender address")

    if resolver is None:
        resolver = GlobalResolver()

    destination_address, endpoints = await resolver.resolve(destination)
    if not endpoints or not destination_address:
        return MsgStatus(
            status=DeliveryStatus.FAILED,
            detail="Failed to resolve destination address",
            destination=destination,
            endpoint="",
            session=None,
        )

    env = Envelope(
        version=1,
        sender=sender_address,
        target=destination_address,
        session=uuid.uuid4(),
        schema_digest=message_schema_digest,
        expires=int(time()) + timeout,
    )
    env.encode_payload(message_body)
    if not is_user_address(sender_address) and isinstance(sender, Identity):
        env.sign(sender.sign_digest)

    response = await send_exchange_envelope(
        envelope=env,
        endpoints=endpoints,
        sync=sync,
    )
    if isinstance(response, Envelope):
        if env.signature is None:
            return response
        json_message = response.decode_payload()
        if response_type:
            return response_type.model_validate_json(json_message)
        return json_message
    return response


async def send_message(
    destination: str,
    message: Model,
    response_type: Optional[Type[Model]] = None,
    sender: Optional[Union[Identity, str]] = None,
    resolver: Optional[Resolver] = None,
    timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
    sync: bool = False,
) -> Union[Model, JsonStr, MsgStatus, Envelope]:
    """
    Standalone function to send a message to an agent.

    Args:
        destination (str): The destination address to send the message to.
        message (Model): The message to be sent.
        response_type (Optional[Type[Model]]): The optional type of the response message.
        sender (Optional[Union[Identity, str]]): The optional sender identity or user address.
        resolver (Optional[Resolver]): The optional resolver for address-to-endpoint resolution.
        timeout (int): The timeout for the message response in seconds. Defaults to 30.
        sync (bool): True if the message is synchronous.

    Returns:
        Union[Model, JsonStr, MsgStatus, Envelope]: On success, if the response type is provided,
        the response message is returned with that type. Otherwise, the JSON message is returned.
        If the sender is a user address, the response envelope is returned.
        On failure, a message status is returned.
    """
    return await send_message_raw(
        destination,
        Model.build_schema_digest(message),
        message.model_dump_json(),
        response_type,
        sender,
        resolver,
        timeout,
        sync,
    )


async def send_sync_message(
    destination: str,
    message: Model,
    response_type: Optional[Type[Model]] = None,
    sender: Optional[Union[Identity, str]] = None,
    resolver: Optional[Resolver] = None,
    timeout: int = DEFAULT_ENVELOPE_TIMEOUT_SECONDS,
) -> Union[Model, JsonStr, MsgStatus, Envelope]:
    """
    Standalone function to send a synchronous message to an agent.

    Args:
        destination (str): The destination address to send the message to.
        message (Model): The message to be sent.
        response_type (Optional[Type[Model]]): The optional type of the response message.
        sender (Optional[Union[Identity, str]]): The optional sender identity or user address.
        resolver (Optional[Resolver]): The optional resolver for address-to-endpoint resolution.
        timeout (int): The timeout for the message response in seconds. Defaults to 30.
        sync (bool): True if the message is synchronous.

    Returns:
        Union[Model, JsonStr, MsgStatus, Envelope]: On success, if the response type is provided,
        the response message is returned with that type. Otherwise, the JSON message is returned.
        If the sender is a user address, the response envelope is returned.
        On failure, a message status is returned.
    """
    return await send_message(
        destination, message, response_type, sender, resolver, timeout, True
    )


def enclose_response(
    message: Model, sender: str, session: UUID4, target: str = ""
) -> JsonStr:
    """
    Enclose a response message within an envelope.

    Args:
        message (Model): The response message to enclose.
        sender (str): The sender's address.
        session (str): The session identifier.
        target (str): The target address.

    Returns:
        str: The JSON representation of the response envelope.
    """
    schema_digest = Model.build_schema_digest(message)
    return enclose_response_raw(
        message.model_dump_json(), schema_digest, sender, session, target
    )


def enclose_response_raw(
    json_message: JsonStr,
    schema_digest: str,
    sender: str,
    session: UUID4,
    target: str = "",
) -> JsonStr:
    """
    Enclose a raw response message within an envelope.

    Args:
        json_message (JsonStr): The JSON-formatted response message to enclose.
        schema_digest (str): The schema digest of the message.
        sender (str): The sender's address.
        session (UUID4): The session identifier.
        target (str): The target address.

    Returns:
        str: The JSON representation of the response envelope.
    """
    response_env = Envelope(
        version=1,
        sender=sender,
        target=target,
        session=session,
        schema_digest=schema_digest,
    )
    response_env.encode_payload(json_message)
    return response_env.model_dump_json()



"""Query Envelopes."""

from typing import Optional, Union

from uagents.communication import MsgStatus, send_sync_message
from uagents.crypto import generate_user_address
from uagents.envelope import Envelope
from uagents.models import Model
from uagents.resolver import Resolver


async def query(
    destination: str,
    message: Model,
    resolver: Optional[Resolver] = None,
    timeout: int = 30,
) -> Union[MsgStatus, Envelope]:
    """
    Query a remote agent with a message and retrieve the response envelope.

    Args:
        destination (str): The destination address of the remote agent.
        message (Model): The message to send.
        resolver (Optional[Resolver], optional): The resolver to use for endpoint resolution.
        Defaults to GlobalResolver.
        timeout (int): The timeout for the query in seconds. Defaults to 30.

    Returns:
        Union[MsgStatus, Envelope]: The response envelope if successful, otherwise MsgStatus.
    """
    response = await send_sync_message(
        destination=destination,
        message=message,
        response_type=None,
        sender=generate_user_address(),
        resolver=resolver,
        timeout=timeout,
    )
    if isinstance(response, (MsgStatus, Envelope)):
        return response
    raise ValueError("Invalid response received.")
